//---------------------------------------------------------------------
// <copyright file="SfxCA.cpp" company="Microsoft">
// Copyright  2005-2007 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
// Main code file for self-extracting MSI managed custom action
// packages. Extracts to temp working directory, hosts the CLR,
// and and invokes an entry point in an extracted module.
//
// Optionally, all of this can be done in a separate process
// with a remote-MSI-API channel hooked up through a named pipe.
//
// Part of the Deployment Tools Foundation project.
// </summary>
//---------------------------------------------------------------------

#include "precomp.h"
#include "EntryPoints.h"
#include "RemoteMsiSession.h"

const bool MANAGED_CAs_OUT_OF_PROC = true;

static HMODULE g_hModule;
static bool g_fRunningOutOfProc = false;

static RemoteMsiSession* g_pRemote = NULL;

// Prototypes for local functions.
// See the function definitions for comments.

void Log(MSIHANDLE hSession, const wchar_t* szMessage, ...);

bool DeleteDirectory(const wchar_t* szDir);

bool ExtractToTempDirectory(MSIHANDLE hSession, HMODULE hModule,
	wchar_t* szTempDir, DWORD cchTempDirBuf);

bool InvokeManagedCustomAction(MSIHANDLE hSession,
	_AppDomain* pAppDomain, const wchar_t* szEntryPoint, int* piResult);

bool LoadCLR(MSIHANDLE hSession, const wchar_t* szVersion, const wchar_t* szConfigFile,
	const wchar_t* szPrimaryAssembly, ICorRuntimeHost** ppHost);

bool CreateAppDomain(MSIHANDLE hSession, ICorRuntimeHost* pHost,
	const wchar_t* szName, const wchar_t* szAppBase,
	const wchar_t* szConfigFile, _AppDomain** ppAppDomain);

bool GetMethod(MSIHANDLE hSession, _AppDomain* pAppDomain,
	const wchar_t* szAssembly, const wchar_t* szClass,
	const wchar_t* szMethod, _MethodInfo** ppCAMethod);

int ExtractCabinet(const wchar_t* szCabFile, const wchar_t* szExtractDir);


/// <summary>
/// Entry-point for the CA DLL when re-launched as a separate process;
/// connects the comm channel for remote MSI APIs, then invokes the
/// managed custom action entry-point.
/// </summary>
/// <remarks>
/// Do not change the parameters or calling-convention: RUNDLL32
/// requires this exact signature.
/// </remarks>
extern "C"
void __stdcall InvokeManagedCustomActionOutOfProc(
	HWND hwnd, HINSTANCE hinst, wchar_t* szCmdLine, int nCmdShow)
{
	g_fRunningOutOfProc = true;

	const wchar_t* szSessionName = szCmdLine;
	MSIHANDLE hSession;
	const wchar_t* szEntryPoint;

	int i;
	for (i = 0; szCmdLine[i] && szCmdLine[i] != L' '; i++);
	if (szCmdLine[i] != L'\0') szCmdLine[i++] = L'\0';
	hSession = _wtoi(szCmdLine + i);

	for (; szCmdLine[i] && szCmdLine[i] != L' '; i++);
	if (szCmdLine[i] != L'\0') szCmdLine[i++] = L'\0';
	szEntryPoint = szCmdLine + i;

	g_pRemote = new RemoteMsiSession(szSessionName, false);
	g_pRemote->Connect();

	int ret = InvokeCustomAction(hSession, NULL, szEntryPoint);

	RemoteMsiSession::RequestData requestData;
    ZeroMemory(&requestData, sizeof(RemoteMsiSession::RequestData));
	requestData.fields[0].vt = VT_I4;
	requestData.fields[0].iValue = ret;
	g_pRemote->SendRequest(RemoteMsiSession::EndSession, &requestData, NULL);
}

/// <summary>
/// Re-launch this CA DLL as a separate process, and setup a comm channel
/// for remote MSI API calls back to this process.
/// </summary>
int InvokeOutOfProcManagedCustomAction(MSIHANDLE hSession, const wchar_t* szEntryPoint)
{
	wchar_t szSessionName[100];
	swprintf_s(szSessionName, 100, L"SfxCA_%d", ::GetTickCount());
	
	RemoteMsiSession remote(szSessionName, true);

	DWORD ret = remote.Connect();
	if (ret != 0)
	{
		Log(hSession, L"Failed to create communication pipe for new CA process. Error code: %d", ret);
		return ERROR_INSTALL_FAILURE;
	}

	ret = remote.ProcessRequests();
	if (ret != 0)
	{
		Log(hSession, L"Failed to open communication pipe for new CA process. Error code: %d", ret);
		return ERROR_INSTALL_FAILURE;
	}

	wchar_t szModule[MAX_PATH];
	GetModuleFileName(g_hModule, szModule, MAX_PATH);

	const wchar_t* rundll32 = L"rundll32.exe";
	const wchar_t* entry = L"zzzInvokeManagedCustomActionOutOfProc";
	wchar_t szCommandLine[1024];
	swprintf_s(szCommandLine, 1024, L"%s \"%s\",%s %s %d %s",
		rundll32, szModule, entry, szSessionName, hSession, szEntryPoint);

	STARTUPINFO si;
	ZeroMemory(&si, sizeof(STARTUPINFO));
	si.cb = sizeof(STARTUPINFO);
	
	PROCESS_INFORMATION pi;
	ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));

	if (!CreateProcess(rundll32, szCommandLine, NULL, NULL, FALSE,
		0, NULL, NULL, &si, &pi))
	{
		DWORD err = GetLastError();
		Log(hSession, L"Failed to create new CA process via RUNDLL32. Error code: %d", err);
		return ERROR_INSTALL_FAILURE;
	}

    DWORD dwWait = WaitForSingleObject(pi.hProcess, INFINITE);
    if (dwWait != WAIT_OBJECT_0)
    {
		DWORD err = GetLastError();
		Log(hSession, L"Failed to wait for CA process. Error code: %d", err);
		return ERROR_INSTALL_FAILURE;
    }

    DWORD dwExitCode;
    BOOL bRet = GetExitCodeProcess(pi.hProcess, &dwExitCode);
    if (!bRet)
    {
		DWORD err = GetLastError();
		Log(hSession, L"Failed to get exit code of CA process. Error code: %d", err);
		return ERROR_INSTALL_FAILURE;
    }
	else if (dwExitCode != 0)
	{
		Log(hSession, L"RUNDLL32 returned error code: %d", dwExitCode);
		return ERROR_INSTALL_FAILURE;
	}

    CloseHandle(pi.hThread);
    CloseHandle(pi.hProcess);

	return remote.ExitCode;
}

/// <summary>
/// Entrypoint for the managed CA proxy (RemotableNativeMethods) to
/// call MSI APIs remotely.
/// </summary>
void __stdcall MsiRemoteInvoke(RemoteMsiSession::RequestId id, RemoteMsiSession::RequestData* pRequest, RemoteMsiSession::RequestData** ppResponse)
{
	if (g_fRunningOutOfProc)
	{
		g_pRemote->SendRequest(id, pRequest, ppResponse);
	}
	else
	{
		*ppResponse = NULL;
	}
}

/// <summary>
/// Invokes a managed custom action from native code by
/// extracting the package to a temporary working directory
/// then hosting the CLR and locating and calling the entrypoint.
/// </summary>
/// <param name="hSession">Handle to the installation session.
/// Passed to custom action entrypoints by the installer engine.</param>
/// <param name="szWorkingDir">Directory containing the CA binaries
/// and the CustomAction.config file defining the entrypoints.
/// This may be NULL, in which case the current module must have
/// a concatenated cabinet containing those files, which will be
/// extracted to a temporary directory.</param>
/// <param name="szEntryPoint">Name of the CA entrypoint to be invoked.
/// This must be either an explicit &quot;AssemblyName!Namespace.Class.Method&quot;
/// string, or a simple name that maps to a full entrypoint definition
/// in CustomAction.config.</param>
/// <returns>The value returned by the managed custom action method,
/// or ERROR_INSTALL_FAILURE if the CA could not be invoked.</returns>
int InvokeCustomAction(MSIHANDLE hSession,
	const wchar_t* szWorkingDir, const wchar_t* szEntryPoint)
{
	if (!g_fRunningOutOfProc && MANAGED_CAs_OUT_OF_PROC && szWorkingDir == NULL)
	{
		return InvokeOutOfProcManagedCustomAction(hSession, szEntryPoint);
	}

	wchar_t szTempDir[MAX_PATH];
	bool fDeleteTemp = false;
	if (szWorkingDir == NULL)
	{
		if (!ExtractToTempDirectory(hSession, g_hModule, szTempDir, MAX_PATH))
		{
			return ERROR_INSTALL_FAILURE;
		}
		szWorkingDir = szTempDir;
		fDeleteTemp = true;
	}

	wchar_t szConfigFilePath[MAX_PATH + 20];
	StringCchCopy(szConfigFilePath, MAX_PATH + 20, szWorkingDir);
	StringCchCat(szConfigFilePath, MAX_PATH + 20, L"\\CustomAction.config");

	const wchar_t* szConfigFile = szConfigFilePath;
	if (!::PathFileExists(szConfigFilePath))
	{
		szConfigFile = NULL;
	}

	wchar_t szWIAssembly[MAX_PATH + 50];
	StringCchCopy(szWIAssembly, MAX_PATH + 50, szWorkingDir);
	StringCchCat(szWIAssembly, MAX_PATH + 50, L"\\Microsoft.Deployment.WindowsInstaller.dll");

	int iResult = ERROR_INSTALL_FAILURE;
	ICorRuntimeHost* pHost;
	if (LoadCLR(hSession, NULL, szConfigFile, szWIAssembly, &pHost))
	{
		_AppDomain* pAppDomain;
		if (CreateAppDomain(hSession, pHost, L"CustomAction", szWorkingDir,
			szConfigFile, &pAppDomain))
		{
			if (!InvokeManagedCustomAction(hSession, pAppDomain, szEntryPoint, &iResult))
			{
				iResult = ERROR_INSTALL_FAILURE;
			}
			HRESULT hr = pHost->UnloadDomain(pAppDomain);
			if (FAILED(hr))
			{
				Log(hSession, L"Failed to unload app domain. Error code 0x%X", hr);
			}
			pAppDomain->Release();
		}

		HRESULT hr = pHost->Stop();
		pHost->Release();
	}

	if (fDeleteTemp)
	{
		DeleteDirectory(szTempDir);
	}
	return iResult;
}

/// <summary>
/// Called by the system when the DLL is loaded.
/// Saves the module handle for later use.
/// </summary>
BOOL WINAPI DllMain(HMODULE hModule, DWORD  dwReason, void* pReserved)
{
	switch (dwReason)
	{
		case DLL_PROCESS_ATTACH:
			g_hModule = hModule;
			break;
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
    return TRUE;
}

/// <summary>
/// Writes a formatted message to the MSI log.
/// Does out-of-proc MSI calls if necessary.
/// </summary>
void Log(MSIHANDLE hSession, const wchar_t* szMessage, ...)
{
	const int LOG_BUFSIZE = 4096;
	wchar_t szBuf[LOG_BUFSIZE];
	va_list args;
	va_start(args, szMessage);
	StringCchVPrintf(szBuf, LOG_BUFSIZE, szMessage, args);

	if (!g_fRunningOutOfProc || g_pRemote == NULL)
	{
		MSIHANDLE hRec = MsiCreateRecord(1);
		MsiRecordSetString(hRec, 0, L"SFXCA: [1]");
		MsiRecordSetString(hRec, 1, szBuf);
		MsiProcessMessage(hSession, INSTALLMESSAGE_INFO, hRec);
		MsiCloseHandle(hRec);
	}
	else
	{
		// Logging is the only remote-MSI operation done from unmanaged code.
		// It's not very convenient here because part of the infrastructure
		// for remote MSI APIs is on the managed side.

		RemoteMsiSession::RequestData req;
		RemoteMsiSession::RequestData* pResp;
		ZeroMemory(&req, sizeof(RemoteMsiSession::RequestData));

		req.fields[0].vt = VT_UI4;
		req.fields[0].uiValue = 1;
		g_pRemote->SendRequest(RemoteMsiSession::MsiCreateRecord, &req, &pResp);
		MSIHANDLE hRec = (MSIHANDLE) pResp->fields[0].iValue;

		req.fields[0].vt = VT_I4;
		req.fields[0].iValue = (int) hRec;
		req.fields[1].vt = VT_UI4;
		req.fields[1].uiValue = 0;
		req.fields[2].vt = VT_LPWSTR;
		req.fields[2].szValue = L"SFXCA: [1]";
		g_pRemote->SendRequest(RemoteMsiSession::MsiRecordSetString, &req, &pResp);

		req.fields[0].vt = VT_I4;
		req.fields[0].iValue = (int) hRec;
		req.fields[1].vt = VT_UI4;
		req.fields[1].uiValue = 1;
		req.fields[2].vt = VT_LPWSTR;
		req.fields[2].szValue = szBuf;
		g_pRemote->SendRequest(RemoteMsiSession::MsiRecordSetString, &req, &pResp);

		req.fields[0].vt = VT_I4;
		req.fields[0].iValue = (int) hSession;
		req.fields[1].vt = VT_I4;
		req.fields[1].iValue = (int) INSTALLMESSAGE_INFO;
		req.fields[2].vt = VT_I4;
		req.fields[2].iValue = (int) hRec;
		g_pRemote->SendRequest(RemoteMsiSession::MsiProcessMessage, &req, &pResp);

		req.fields[0].vt = VT_I4;
		req.fields[0].iValue = (int) hRec;
		g_pRemote->SendRequest(RemoteMsiSession::MsiCloseHandle, &req, &pResp);
	}
}

/// <summary>
/// Deletes a directory, including all files and subdirectories.
/// </summary>
/// <param name="szDir">Path to the directory to delete,
/// not including a trailing backslash.</param>
/// <returns>True if the directory was successfully deleted, or false
/// if the deletion failed (most likely because some files were locked).
/// </returns>
bool DeleteDirectory(const wchar_t* szDir)
{
	size_t cchDir = wcslen(szDir);
	size_t cchPathBuf = cchDir + 3 + MAX_PATH;
	wchar_t* szPath = (wchar_t*) _alloca(cchPathBuf * sizeof(wchar_t));
	if (szPath == NULL) return false;
	StringCchCopy(szPath, cchPathBuf, szDir);
	StringCchCat(szPath, cchPathBuf, L"\\*");
	WIN32_FIND_DATA fd;
	HANDLE hSearch = FindFirstFile(szPath, &fd);
	while (hSearch != INVALID_HANDLE_VALUE)
	{
		StringCchCopy(szPath + cchDir + 1, cchPathBuf - (cchDir + 1), fd.cFileName);
		if ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
		{
			if (wcscmp(fd.cFileName, L".") != 0 && wcscmp(fd.cFileName, L"..") != 0)
			{
				DeleteDirectory(szPath);
			}
		}
		else
		{
			DeleteFile(szPath);
		}
		if (!FindNextFile(hSearch, &fd))
		{
			FindClose(hSearch);
			hSearch = INVALID_HANDLE_VALUE;
		}
	}
	return RemoveDirectory(szDir) != 0;
}

bool DirectoryExists(const wchar_t* szDir)
{
    if (szDir != NULL)
    {
        DWORD dwAttrs = GetFileAttributes(szDir);
        if (dwAttrs != -1 && (dwAttrs & FILE_ATTRIBUTE_DIRECTORY) != 0)
        {
            return true;
        }
    }
    return false;
}

/// <summary>
/// Extracts a cabinet that is concatenated to a module
/// to a new temporary directory.
/// </summary>
/// <param name="hSession">Handle to the installer session,
/// used just for logging.</param>
/// <param name="hModule">Module that has the concatenated cabinet.</param>
/// <param name="szTempDir">Buffer for returning the path of the
/// created temp directory.</param>
/// <param name="cchTempDirBuf">Size in characters of the buffer.
/// <returns>True if the files were extracted, or false if the
/// buffer was too small or the directory could not be created
/// or the extraction failed for some other reason.</returns>
bool ExtractToTempDirectory(MSIHANDLE hSession, HMODULE hModule,
	wchar_t* szTempDir, DWORD cchTempDirBuf)
{
	wchar_t szModule[MAX_PATH];
	DWORD cchCopied = GetModuleFileName(hModule, szModule, MAX_PATH - 1);
	if (cchCopied == 0)
	{
		Log(hSession, L"Failed to get module path. Error code %d.", GetLastError());
		return false;
	}
	else if (cchCopied == MAX_PATH - 1)
	{
		Log(hSession, L"Failed to get module path -- path is too long.");
		return false;
	}

	if (szTempDir == NULL || cchTempDirBuf < wcslen(szModule) + 1)
	{
		Log(hSession, L"Temp directory buffer is NULL or too small.");
		return false;
	}
	StringCchCopy(szTempDir, cchTempDirBuf, szModule);
	StringCchCat(szTempDir, cchTempDirBuf, L"-");

    DWORD cchTempDir = (DWORD) wcslen(szTempDir);
    for (int i = 0; DirectoryExists(szTempDir); i++)
    {
        swprintf_s(szTempDir + cchTempDir, cchTempDirBuf - cchTempDir, L"%d", i);
    }

	if (!CreateDirectory(szTempDir, NULL))
	{
		cchCopied = GetTempPath(cchTempDirBuf, szTempDir);
		if (cchCopied == 0 || cchCopied >= cchTempDirBuf)
		{
			Log(hSession, L"Failed to get temp directory. Error code %d", GetLastError());
			return false;
		}

		wchar_t* szModuleName = wcsrchr(szModule, L'\\');
		if (szModuleName == NULL) szModuleName = szModule;
		else szModuleName = szModuleName + 1;
		StringCchCat(szTempDir, cchTempDirBuf, szModuleName);
		StringCchCat(szTempDir, cchTempDirBuf, L"-");

        cchTempDir = (DWORD) wcslen(szTempDir);
        for (int i = 0; DirectoryExists(szTempDir); i++)
        {
            swprintf_s(szTempDir + cchTempDir, cchTempDirBuf - cchTempDir, L"%d", i);
        }

		if (!CreateDirectory(szTempDir, NULL))
		{
			Log(hSession, L"Failed to create temp directory. Error code %d", GetLastError());
			return false;
		}
	}

	Log(hSession, L"Extracting custom action to temporary directory: %s\\", szTempDir);
	int err = ExtractCabinet(szModule, szTempDir);
    if (err != 0)
	{
      	Log(hSession, L"Failed to extract to temporary directory. Cabinet error code %d.", err);
		DeleteDirectory(szTempDir);
		return false;
	}
	return true;
}

/// <summary>
/// Loads and invokes the managed portion of the proxy.
/// </summary>
/// <param name="hSession">Handle to the installer session,
/// used for logging errors and to be passed on to the custom action.</param>
/// <param name="pAppDomain">AppDomain which has its application
/// base set to the CA working directory.</param>
/// <param name="szEntryPoint">Name of the CA entrypoint to be invoked.
/// This must be either an explicit &quot;AssemblyName!Namespace.Class.Method&quot;
/// string, or a simple name that maps to a full entrypoint definition
/// in CustomAction.config.</param>
/// <param name="piResult">Return value of the invoked custom
/// action method.</param>
/// <returns>True if the managed proxy was invoked successfully,
/// false if there was some error. Note the custom action itself may
/// return an error via piResult while this method still returns true
/// since the invocation was successful.</returns>
bool InvokeManagedCustomAction(MSIHANDLE hSession, _AppDomain* pAppDomain,
	const wchar_t* szEntryPoint, int* piResult)
{
	const wchar_t* szMsiAssemblyName   = L"Microsoft.Deployment.WindowsInstaller";
	const wchar_t* szMsiCAProxyClass   = L"Microsoft.Deployment.WindowsInstaller.CustomActionProxy";
	const wchar_t* szMsiCAInvokeMethod = L"InvokeCustomAction";
	
	_MethodInfo* pCAInvokeMethod;
	if (!GetMethod(hSession, pAppDomain, szMsiAssemblyName,
		szMsiCAProxyClass, szMsiCAInvokeMethod, &pCAInvokeMethod))
	{
		return false;
	}

	HRESULT hr;
	VARIANT vNull;
	vNull.vt = VT_EMPTY;
	SAFEARRAYBOUND sabound;
	sabound.lLbound = 0;
	sabound.cElements = 3;
	SAFEARRAY* saArgs = SafeArrayCreate(VT_VARIANT, 1, &sabound);
	VARIANT vSessionHandle;
	vSessionHandle.vt = VT_I4;
	vSessionHandle.intVal = hSession;
	LONG index = 0;
	hr = SafeArrayPutElement(saArgs, &index, &vSessionHandle);
	if (FAILED(hr)) goto LExit;
	VARIANT vEntryPoint;
	vEntryPoint.vt = VT_BSTR;
	vEntryPoint.bstrVal = SysAllocString(szEntryPoint);
	if (vEntryPoint.bstrVal == NULL)
	{
		hr = E_OUTOFMEMORY;
		goto LExit;
	}
	index = 1;
	hr = SafeArrayPutElement(saArgs, &index, &vEntryPoint);
	if (FAILED(hr)) goto LExit;
	VARIANT vRemotingFunctionPtr;
	vRemotingFunctionPtr.vt =  VT_I8;
	vRemotingFunctionPtr.llVal = (LONGLONG) (g_fRunningOutOfProc ? MsiRemoteInvoke : NULL);
	index = 2;
	hr = SafeArrayPutElement(saArgs, &index, &vRemotingFunctionPtr);
	if (FAILED(hr)) goto LExit;
	
	VARIANT vResult;
	hr = pCAInvokeMethod->Invoke_3(vNull, saArgs, &vResult);

LExit:
	SafeArrayDestroy(saArgs);
	pCAInvokeMethod->Release();
	
	if (FAILED(hr))
	{
		Log(hSession, L"Failed to invoke custom action method. Error code 0x%X", hr);
		return false;
	}

	*piResult = vResult.intVal;
	return true;
}

